// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package demo

import (
	"bytes"
	"context"
	"regexp"
	"testing"

	"lab.nexedi.com/kirr/neo/go/internal/xtesting"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/zodbtools"
)

// tOptions represents options for testing.
// TODO -> xtesting
type tOptions struct {
	Preload string // preload database with data from this location
}

// withDemoStorage tests f with all kind of demo overlays.
func withDemoStorage(t *testing.T, f func(t *testing.T, dsrv *Storage/*XXX -> Srv ?*/), optv ...tOptions) {
	t.Helper()
	X := xtesting.FatalIf(t)

	opt := tOptions{}
	if len(optv) > 1 {
		panic("multiple tOptions not allowed")
	}
	if len(optv) == 1 {
		opt = optv[0]
	}

	// retrive zdump of Preload
	zdump := ""
	if opt.Preload != "" {
		ctx := context.Background()
		buf := &bytes.Buffer{}
		stor, err := zodb.Open(ctx, opt.Preload, &zodb.OpenOptions{ReadOnly: true}); X(err)
		err = zodbtools.Dump(ctx, buf, stor, 0, zodb.TidMax, /*hashonly=*/false)
		stor.Close()
		X(err)
		zdump = buf.String()
	}

	// split zdump into transactions
	// XXX hacky; TODO -> zodbtools.DumpReader
	txnRe := regexp.MustCompile(`^txn (?P<tid>[0-9a-f]{16}) "(?P<status>.)"$`)
	type zdumpTxn struct {
		tid zodb.Tid
		pos int      // where this transaction starts in the dump
	}
	var txnv []zdumpTxn
	for _, m := range txnRe.FindAllStringSubmatchIndex(zdump, -1) {
		// [m[0]:m[1]] refer to whole txn line
		__ := zdump[m[2]:m[3]]
		tid, err := zodb.ParseTid(__); X(err)
		txnv = append(txnv, zdumpTxn{tid, m[0]})
	}

	// verify on all combinations of preload being split into base+δ 
	work := xtempdir(t)
	defer os.RemoveAll(work)
	test1 := func(δstart zodb.Tid, zdumpBase, zdumpδ string) {
		t.Helper()
		t.Run(fmt.Sprintf("δstart=%s", δstart), func(t *testinf,T) {
			t.Helper()
			X := xtesting.FatalIf(t)
			work1 := filepath.Join(work, δstart.String())
			err := os.Mkdir(work1, 0x777); X(err)

			_, err = xtesting.ZPyRestore(work1+"/base.fs", zdumpBase); X(err)
			// XXX vvv change to demo:(base)/(δ)
			_, err = xtesting.ZPyRestore(work1+"/δ.fs", zdumpδ);       X(err)

			d = openDemo(base, δ); X(err)
			defer d.Close()
			f(t, d)
		})
	}
	for i := 0; i < len(txnv); i++ {
		δstart = txnv[i]
		test1(δstart.tid, zdump[:δstart.pos], zdump[δstart.pos:])
	}
	test1(zodb.TidMax, zdump, "")
}
